﻿using System;
// demo of Fisher's Linear Discriminate Analysis (LDA)
// for binary classification
// note: a different LDA is Latent Dirichlet Allocation 

namespace LinearDiscriminate
{
  class LinearDiscriminateProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin linear discriminate analysis (LDA) demo\n");
      Console.WriteLine("Goal is to predict Politic (liberal = 0 or conservative = 1) ");
      Console.WriteLine("from person's Age and Income\n");

      double[][] data = new double[8][];
      data[0] = new double[] { 1.0, 4.0, 0 };
      data[1] = new double[] { 2.0, 2.0, 0 };
      data[2] = new double[] { 3.0, 3.0, 0 };

      data[3] = new double[] { 6.0, 6.0, 1 };
      data[4] = new double[] { 7.0, 7.0, 1 };
      data[5] = new double[] { 8.0, 9.0, 1 };
      data[6] = new double[] { 8.0, 7.0, 1 };
      data[7] = new double[] { 9.0, 8.0, 1 };

      Console.WriteLine("Normalized and encoded training data:\n");
      Console.WriteLine(" Age   Income  Politic");
      Console.WriteLine("----------------------");
      ShowMatrix(data, 2, true);

      Console.WriteLine("\nCalculating the discriminate w");
      double[][] w = Discriminate(data, true); // calls ScatterWithin which calls Scatter
      Console.WriteLine("Done calculating discriminate. Discriminate is: \n");
      ShowMatrix(w, 6, true);
     
      double[] x = new double[] { 4.0, 5.0 };
      //double[] unknown = new double[] { 5.6, 5.0 };
      Console.WriteLine("Predicting class for Age Income =");
      ShowVector(x, 1);
      int c = Prediction(data, x, w); 
      Console.Write("\nPredicted class: " + c);
      if (c == 0)
        Console.WriteLine(" (liberal)");
      else
        Console.WriteLine(" (conservative)");

      // 2. pooled covariance approach to compute Discriminate
      //double[][] pooled = Pooled(data, false);
      //double[][] pooledInv = MatrixInverse(pooled);
      //double[][] M0 = Mean(data, 0);
      //double[][] M1 = Mean(data, 1);
      //double[][] diff = MatrixSubtract(M0, M1);
      //double[][] prod = MatrixProduct(pooledInv, diff);
      //double[][] pooledW = Unitize(prod); 
      //ShowMatrix(pooledW, 6, true); // exactly same as non-pooled scatter
      //Console.WriteLine("\nDiscriminate calculated using pooled covariance:");
      //ShowMatrix(pooledW, 4, true);

      // Sb (between-class scatter) not explicitly used 
      //Two different equations for Sb
      //See https://www.ics.uci.edu/~welling/teaching/KernelsICS273B/Fisher-LDA.pdf

      Console.WriteLine("\nEnd LDA binary classification demo\n");
      Console.ReadLine();
    } // Main

    static int Prediction(double[][] data, double[] x, double[][] w)
    {
      // item to predict, x, is a normal array
      // convert array to 1-col matrix then call regular Prediction()
      double[][] xm = MatrixFromVector(x); // x to col matrix
      return Prediction(data, xm, w);
    }

    static int Prediction(double[][] data, double[][] x, double[][] w)
    {
      // item to predict, x, is a 1-col matrix (column vector)
      int dim = data[0].Length - 1; // at least 2
      double[][] wT = MatrixTranspose(w); // 1 x d
      double[][] m0 = Mean(data, 0); // d x 1
      double[][] m1 = Mean(data, 1); // d x 1
      double[][] m = MatrixAdd(m0, m1); // d x 1
      m = MatrixProduct(m, 0.5); // d x 1 
 
      // threshold constant c (distance from origin)
      double[][] tc = MatrixProduct(wT, m); // ((1xd)(dx1) = 1 x 1
      // distance from origin
      double[][] wTx = MatrixProduct(wT, x); // (1xd)(dx1) = 1 x 1

      double adjust = 0.0;

      // optional weighting by probabilities:
      //int n0 = 0; int n1 = 0; // number each class
      //for (int i = 0; i < data.Length; ++i)
      //{
      //  int label = (int)data[i][dim];
      //  if (label == 0)
      //    ++n0;
      //  else
      //    ++n1;
      //}
      //double p0 = (n0 * 1.0) / data.Length;
      //double p1 = (n1 * 1.0) / data.Length;
      //double adjust = Math.Log(p0 / p1);

      if (wTx[0][0] - adjust > tc[0][0]) // a bit tricky
        return 0;
      else
        return 1;
    }

    static double[][] Mean(double[][] data, int c)
    {
      // return mean of class c (0 or 1) as column vector
      int dim = data[0].Length - 1;
      double[][] result = MatrixCreate(dim, 1);
      int ct = 0;
      
      for (int i = 0; i < data.Length; ++i)
      {
        int label = (int)data[i][dim];
        if (label == c)
        {
          for (int j = 0; j < dim; ++j)
            result[j][0] += data[i][j];
          ++ct;
        }
      } // i

      for (int i = 0; i < dim; ++i)
        result[i][0] /= ct;

      return result;
    } // Mean

    static double[][] Discriminate(double[][] data, bool unitize)
    {
      // returns the w vector as a column 
      // calls helper ScatterWithin which calls Scatter
      double[][] mean0 = Mean(data, 0);
      double[][] mean1 = Mean(data, 1);
      Console.WriteLine("\nClass means: \n");
      ShowMatrix(mean0, 2, true);
      ShowMatrix(mean1, 2, true);

      double[][] Sw = ScatterWithin(data); // sum of S0 and S1
      Console.WriteLine("The within-class combined scatter matrix Sw: \n");
      ShowMatrix(Sw, 4, true);

      double[][] SwInv = MatrixInverse(Sw);
      
      double[][] diff = MatrixSubtract(mean0, mean1);
      double[][] w = MatrixProduct(SwInv, diff);

      if (unitize == true)
        return Unitize(w);
      else
        return w;
    }

    static double[][] ScatterWithin(double[][] data)
    {
      // Sw = within class scatter = S0 + S1
      double[][] S0 = Scatter(data, 0);
      double[][] S1 = Scatter(data, 1);
      Console.WriteLine("Scatter matrices S0 S1: \n");
      ShowMatrix(S0, 4, true);
      ShowMatrix(S1, 4, true);
      double[][] Sw = MatrixAdd(S0, S1);
      return Sw;
    }

    static double[][] Scatter(double[][] data, int c)
    {
      // scatter matrix of class c
      // [Sw (within class) is sum of class scatters]
      int dim = data[0].Length - 1;
      double[][] mean = Mean(data, c); // mean as col vector
      double[][] result = MatrixCreate(dim, dim); // d x d
      for (int i = 0; i < data.Length; ++i)
      {
        int label = (int)data[i][dim]; // 0 or 1
        if (label == c)
        {
          double[][] x = MatrixCreate(dim, 1); // d x 1
          for (int j = 0; j < dim; ++j)
            x[j][0] = data[i][j];
          double[][] diff = MatrixSubtract(x, mean); // d x 1
          double[][] diffT = MatrixTranspose(diff);  // 1 x d
          double[][] prod = MatrixProduct(diff, diffT); // d x d

          result = MatrixAdd(result, prod); // accumulate
        }
      } // i
      return result;
    } // Scatter


    static double[][] Unitize(double[][] vector)
    {
      // return column vector scaled to unit length
      if (vector[0].Length != 1)
        throw new Exception("Not column vector");

      int len = vector.Length;
      double[][] result = MatrixCreate(len, 1);
      double sum = 0.0;
      for (int i = 0; i < len; ++i)
        sum += vector[i][0] * vector[i][0];
      sum = Math.Sqrt(sum); // check if 0
      for (int i = 0; i < len; ++i)
        result[i][0] = vector[i][0] / sum;
      return result;
    }

    // ============================================

    //static double[][] Covariance(double[][] data, int c, bool sample)
    //{
    //  // covariance is essentially Scatter divided by number items
    //  int n = 0;
    //  int dim = data[0].Length - 1;
    //  for (int i = 0; i < data.Length; ++i)
    //  {
    //    int label = (int)data[i][dim];
    //    if (label == c)
    //      ++n;
    //  }
    //  double[][] s = Scatter(data, c);
    //  double[][] result = MatrixDuplicate(s);
    //  for (int i = 0; i < dim; ++i)
    //  {
    //    for (int j = 0; j < dim; ++j)
    //    {
    //      if (sample == true)
    //        result[i][j] /= n - 1;
    //      else
    //        result[i][j] /= n;
    //    }
    //  }
    //  return result;
    //} // Covariance

    //static double[][] Pooled(double[][] data, bool sample)
    //{
    //  // pooled covariance: (n1*C1 + n2*C2) / (n1 + n2)
    //  int n0 = 0; int n1 = 0;
    //  int dim = data[0].Length - 1;
    //  for (int i = 0; i < data.Length; ++i)
    //  {
    //    int label = (int)data[i][dim]; // 0 or 1
    //    if (label == 0) ++n0; else ++n1;
    //  }

    //  double[][] C0 = Covariance(data, 0, sample);
    //  double[][] C1 = Covariance(data, 1, sample);

    //  double[][] result = MatrixCreate(dim, dim);
    //  for (int i = 0; i < dim; ++i)
    //    for (int j = 0; j < dim; ++j)
    //      result[i][j] = (n0 * C0[i][j] + n1 * C1[i][j]) / (n0 + n1);
    //  return result;
    //} // Pooled

    // -------------------------------------------------------------
    // Matrix methods below
    // -------------------------------------------------------------

    static double[][] MatrixInverse(double[][] matrix)
    {
      int n = matrix.Length;
      double[][] result = MatrixDuplicate(matrix);

      int[] perm;
      int toggle;
      double[][] lum = MatrixDecompose(matrix, out perm, out toggle);
      if (lum == null)
        throw new Exception("Unable to compute inverse");

      double[] b = new double[n];
      for (int i = 0; i < n; ++i)
      {
        for (int j = 0; j < n; ++j)
        {
          if (i == perm[j])
            b[j] = 1.0;
          else
            b[j] = 0.0;
        }

        double[] x = HelperSolve(lum, b); // use decomposition

        for (int j = 0; j < n; ++j)
          result[j][i] = x[j];
      }
      return result;
    }

    // -------------------------------------------------------------

    static double[] HelperSolve(double[][] luMatrix, double[] b)
    {
      // before calling this helper, permute b using the perm array
      // from MatrixDecompose that generated luMatrix
      int n = luMatrix.Length;
      double[] x = new double[n];
      b.CopyTo(x, 0);

      for (int i = 1; i < n; ++i)
      {
        double sum = x[i];
        for (int j = 0; j < i; ++j)
          sum -= luMatrix[i][j] * x[j];
        x[i] = sum;
      }

      x[n - 1] /= luMatrix[n - 1][n - 1];
      for (int i = n - 2; i >= 0; --i)
      {
        double sum = x[i];
        for (int j = i + 1; j < n; ++j)
          sum -= luMatrix[i][j] * x[j];
        x[i] = sum / luMatrix[i][i];
      }

      return x;
    }

    // -------------------------------------------------------------

    static double[][] MatrixDuplicate(double[][] matrix)
    {
      // allocates/creates a duplicate of a matrix
      double[][] result = MatrixCreate(matrix.Length, matrix[0].Length);
      for (int i = 0; i < matrix.Length; ++i) // copy the values
        for (int j = 0; j < matrix[i].Length; ++j)
          result[i][j] = matrix[i][j];
      return result;
    }

    // -------------------------------------------------------------

    static double[][] MatrixDecompose(double[][] matrix, out int[] perm,
      out int toggle)
    {
      // Doolittle LUP decomposition with partial pivoting.
      // returns: result is L (with 1s on diagonal) and U;
      // perm holds row permutations; toggle is +1 or -1 (even or odd)
      int rows = matrix.Length;
      int cols = matrix[0].Length;
      if (rows != cols)
        throw new Exception("Non-square mattrix");

      int n = rows; // convenience

      double[][] result = MatrixDuplicate(matrix); // 

      perm = new int[n]; // set up row permutation result
      for (int i = 0; i < n; ++i) { perm[i] = i; }

      toggle = 1; // toggle tracks row swaps

      for (int j = 0; j < n - 1; ++j) // each column
      {
        double colMax = Math.Abs(result[j][j]);
        int pRow = j;
        //for (int i = j + 1; i < n; ++i) // deprecated
        //{
        //  if (result[i][j] > colMax)
        //  {
        //    colMax = result[i][j];
        //    pRow = i;
        //  }
        //}

        for (int i = j + 1; i < n; ++i) // reader Matt V needed this:
        {
          if (Math.Abs(result[i][j]) > colMax)
          {
            colMax = Math.Abs(result[i][j]);
            pRow = i;
          }
        }
        // Not sure if this approach is needed always, or not.

        if (pRow != j) // if largest value not on pivot, swap rows
        {
          double[] rowPtr = result[pRow];
          result[pRow] = result[j];
          result[j] = rowPtr;

          int tmp = perm[pRow]; // and swap perm info
          perm[pRow] = perm[j];
          perm[j] = tmp;

          toggle = -toggle; // adjust the row-swap toggle
        }

        // -------------------------------------------------------------
        // This part added later (not in original code) 
        // and replaces the 'return null' below.
        // if there is a 0 on the diagonal, find a good row 
        // from i = j+1 down that doesn't have
        // a 0 in column j, and swap that good row with row j

        if (result[j][j] == 0.0)
        {
          // find a good row to swap
          int goodRow = -1;
          for (int row = j + 1; row < n; ++row)
          {
            if (result[row][j] != 0.0)
              goodRow = row;
          }

          if (goodRow == -1)
            throw new Exception("Cannot use Doolittle's method");

          // swap rows so 0.0 no longer on diagonal
          double[] rowPtr = result[goodRow];
          result[goodRow] = result[j];
          result[j] = rowPtr;

          int tmp = perm[goodRow]; // and swap perm info
          perm[goodRow] = perm[j];
          perm[j] = tmp;

          toggle = -toggle; // adjust the row-swap toggle
        }
        // -------------------------------------------------------------

        //if (Math.Abs(result[j][j]) < 1.0E-20) // deprecated
        //  return null; // consider a throw

        for (int i = j + 1; i < n; ++i)
        {
          result[i][j] /= result[j][j];
          for (int k = j + 1; k < n; ++k)
          {
            result[i][k] -= result[i][j] * result[j][k];
          }
        }

      } // main j column loop

      return result;
    } // MatrixDecompose

    // -------------------------------------------------------------

    static double[][] MatrixProduct(double[][] matrixA, double[][] matrixB)
    {
      int aRows = matrixA.Length; int aCols = matrixA[0].Length;
      int bRows = matrixB.Length; int bCols = matrixB[0].Length;
      if (aCols != bRows)
        throw new Exception("Non-conformable matrices in MatrixProduct");

      double[][] result = MatrixCreate(aRows, bCols);

      for (int i = 0; i < aRows; ++i) // each row of A
        for (int j = 0; j < bCols; ++j) // each col of B
          for (int k = 0; k < aCols; ++k) // could use k < bRows
            result[i][j] += matrixA[i][k] * matrixB[k][j];

      //Parallel.For(0, aRows, i =>
      //  {
      //    for (int j = 0; j < bCols; ++j) // each col of B
      //      for (int k = 0; k < aCols; ++k) // could use k < bRows
      //        result[i][j] += matrixA[i][k] * matrixB[k][j];
      //  }
      //);

      return result;
    }

    static double[][] MatrixProduct(double[][] m, double x)
    {
      // multiple all cells in m by scalar x
      double[][] result = MatrixDuplicate(m); // copy
      for (int i = 0; i < result.Length; ++i)
        for (int j = 0; j < result[i].Length; ++j)
          result[i][j] *= x;
      return result;
    }

    static double[][] MatrixAdd(double[][] a, double[][] b)
    {
      // return a-b
      int rows = a.Length; int cols = a[0].Length;
      double[][] result = MatrixCreate(rows, cols);
      for (int i = 0; i < rows; ++i)
        for (int j = 0; j < cols; ++j)
          result[i][j] = a[i][j] + b[i][j];
      return result;
    }

    static double[][] MatrixSubtract(double[][] a, double[][] b)
    {
      // return a-b
      int rows = a.Length; int cols = a[0].Length;
      double[][] result = MatrixCreate(rows, cols);
      for (int i = 0; i < rows; ++i)
        for (int j = 0; j < cols; ++j)
          result[i][j] = a[i][j] - b[i][j];
      return result;
    }

    static double[][] MatrixTranspose(double[][] matrix)
    {
      int rows = matrix.Length;
      int cols = matrix[0].Length;
      double[][] result = MatrixCreate(cols, rows); // note indexing
      for (int i = 0; i < rows; ++i)
      {
        for (int j = 0; j < cols; ++j)
        {
          result[j][i] = matrix[i][j];
        }
      }
      return result;
    } // TransposeMatrix

    static double[][] MatrixFromVector(double[] vector)
    {
      // return a column vector-matrix
      int len = vector.Length;
      double[][] result = MatrixCreate(len, 1); // 1 colum
      for (int i = 0; i < len; ++i)
        result[i][0] = vector[i];
      return result;
    }

    static double[][] MatrixCreate(int rows, int cols)
    {
      // allocates/creates a matrix initialized to all 0.0
      // do error checking here
      double[][] result = new double[rows][];
      for (int i = 0; i < rows; ++i)
        result[i] = new double[cols];
      return result;
    }

    static void ShowMatrix(double[][] m, int dec, bool newl)
    {
      for (int i = 0; i < m.Length; ++i)
      {
        for (int j = 0; j < m[i].Length; ++j)
        {
          if (m[i][j] >= 0.0) Console.Write(" "); // '+'
          Console.Write(m[i][j].ToString("F" + dec) + "  ");
        }
        Console.WriteLine("");
      }
      if (newl) Console.WriteLine("");
    }

    static void ShowVector(double[] v, int dec)
    {
      for (int i = 0; i < v.Length; ++i)
        Console.Write(v[i].ToString("F" + dec) + "  ");
      Console.WriteLine("");
    }

  } // Program clss

} // ns


// R language analysis for validation of C# program
//> library(MASS)
//>
//> data1 = read.table(file.choose(), header=T, sep=',')
//>
//> data1
//  Age Income Politic
//1   1      4       A
//2   2      2       A
//3   3      3       A
//4   6      6       B
//5   7      7       B
//6   8      9       B
//7   8      7       B
//8   9      8       B
//>
//> data1.lda <- lda(formula = Politic ~ Age + Income, data = data1)
//>
//> data1.lda

//Prior probabilities of groups:
//    A     B
//0.375 0.625

//Group means:
//  Age Income
//A 2.0    3.0
//B 7.6    7.4

//Coefficients of linear discriminants:
//             LD1
//Age    0.6859669
//Income 0.3919811
//>

// Show the C# w vector is equivalent to the R result
//> v = c(0.6859669, 0.3919811)
//> len = sqrt(v[1]^2 + v[2]^2)
//> w = v / len
//> w
//[1] 0.8682431 0.4961390
//>

// Make a prediction using R
//> unknown <- data.frame(Age=4,Income=5,Politic='?')
//> predict(data1.lda, unknown)
//          A          B
//1 0.9516662 0.04833379

// a few LDA resources
//See http://www.cbcb.umd.edu/~hcorrada/PracticalML/pdf/lectures/classification.pdf
//See http://sites.stat.psu.edu/~jiali/course/stat597e/notes2/lda.pdf
//See http://www.saedsayad.com/lda.htm